@@ -19,7 +19,7 @@ from sqlalchemy.ext.declarative import declarative_base
 from sqlalchemy.orm import backref, relationship
 
 # Zato
-from zato.common import SCHEDULER_JOB_TYPE
+from zato.common import INVOCATION_TARGET, MISC, MSG_PATTERN_TYPE, SCHEDULER_JOB_TYPE
 from zato.common.odb import AMQP_DEFAULT_PRIORITY, WMQ_DEFAULT_PRIORITY
 
 Base = declarative_base()
@@ -44,7 +44,7 @@ class ZatoInstallState(Base):
     """
     __tablename__ = 'install_state'
 
-    id = Column(Integer,  Sequence('install_state_seq'), primary_key=True)
+    id = Column(Integer, Sequence('install_state_seq'), primary_key=True)
     version = Column(Integer, unique=True, nullable=False)
     install_time = Column(DateTime(), nullable=False)
     source_host = Column(String(200), nullable=False)
@@ -63,7 +63,7 @@ class Cluster(Base):
     """
     __tablename__ = 'cluster'
 
-    id = Column(Integer,  Sequence('cluster_id_seq'), primary_key=True)
+    id = Column(Integer, Sequence('cluster_id_seq'), primary_key=True)
     name = Column(String(200), unique=True, nullable=False)
     description = Column(String(1000), nullable=True)
     odb_type = Column(String(30), nullable=False)
@@ -82,7 +82,7 @@ class Cluster(Base):
 
     def __init__(self, id=None, name=None, description=None, odb_type=None,
                  odb_host=None, odb_port=None, odb_user=None, odb_db_name=None,
-                 odb_schema=None, broker_host=None, 
+                 odb_schema=None, broker_host=None,
                  broker_port=None, lb_host=None, lb_port=None,
                  lb_agent_port=None, cw_srv_id=None, cw_srv_keep_alive_dt=None):
         self.id = id
@@ -111,10 +111,10 @@ class Server(Base):
     __tablename__ = 'server'
     __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
 
-    id = Column(Integer,  Sequence('server_id_seq'), primary_key=True)
+    id = Column(Integer, Sequence('server_id_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     host = Column(String(400), nullable=True)
-    
+
     bind_host = Column(String(400), nullable=True)
     bind_port = Column(Integer(), nullable=True)
 
@@ -123,7 +123,7 @@ class Server(Base):
     last_join_status = Column(String(40), nullable=True)
     last_join_mod_date = Column(DateTime(), nullable=True)
     last_join_mod_by = Column(String(200), nullable=True)
-    
+
     # Whether the server's up or not
     up_status = Column(String(40), nullable=True)
     up_mod_date = Column(DateTime(), nullable=True)
@@ -155,29 +155,30 @@ class SecurityBase(Base):
     """ A base class for all the security definitions.
     """
     __tablename__ = 'sec_base'
-    __table_args__ = (UniqueConstraint('cluster_id', 'name'), {})
+    __table_args__ = (UniqueConstraint('cluster_id', 'name'),
+        UniqueConstraint('cluster_id', 'username', 'sec_type'), {})
     __mapper_args__ = {'polymorphic_on': 'sec_type'}
-    
-    id = Column(Integer,  Sequence('sec_base_seq'), primary_key=True)
+
+    id = Column(Integer, Sequence('sec_base_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
-    
+
     # It's nullable because TechnicalAccount doesn't use usernames
-    username = Column(String(200), nullable=True) 
-    
-    password = Column(String(64), nullable=False)
+    username = Column(String(200), nullable=True)
+
+    password = Column(String(64), nullable=True)
     password_type = Column(String(45), nullable=True)
     is_active = Column(Boolean(), nullable=False)
     sec_type = Column(String(45), nullable=False)
-    
+
     cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
     cluster = relationship(Cluster, backref=backref('http_basic_auth_list', order_by=name, cascade='all, delete, delete-orphan'))
 
 class HTTPBasicAuth(SecurityBase):
     """ An HTTP Basic Auth definition.
-    """    
+    """
     __tablename__ = 'sec_basic_auth'
     __mapper_args__ = {'polymorphic_identity': 'basic_auth'}
-    
+
     id = Column(Integer, ForeignKey('sec_base.id'), primary_key=True)
     realm = Column(String(200), nullable=False)
 
@@ -196,7 +197,7 @@ class WSSDefinition(SecurityBase):
     """
     __tablename__ = 'sec_wss_def'
     __mapper_args__ = {'polymorphic_identity':'wss'}
-    
+
     id = Column(Integer, ForeignKey('sec_base.id'), primary_key=True)
     reject_empty_nonce_creat = Column(Boolean(), nullable=False)
     reject_stale_tokens = Column(Boolean(), nullable=True)
@@ -226,7 +227,7 @@ class TechnicalAccount(SecurityBase):
     """
     __tablename__ = 'sec_tech_acc'
     __mapper_args__ = {'polymorphic_identity':'tech_acc'}
-    
+
     id = Column(Integer, ForeignKey('sec_base.id'), primary_key=True)
     salt = Column(String(32), nullable=False)
 
@@ -237,7 +238,34 @@ class TechnicalAccount(SecurityBase):
         self.password = password
         self.salt = salt
         self.cluster = cluster
-        
+
+    def to_json(self):
+        return to_json(self)
+
+class OAuth(SecurityBase):
+    """ New in 1.2: Stores OAuth credentials.
+    """
+    __tablename__ = 'sec_oauth'
+    __mapper_args__ = {'polymorphic_identity':'oauth'}
+
+    id = Column(Integer, ForeignKey('sec_base.id'), primary_key=True)
+    proto_version = Column(String(32), nullable=False)
+    sig_method = Column(String(32), nullable=False) # HMAC-SHA1 or PLAINTEXT
+    max_nonce_log = Column(Integer(), nullable=False)
+
+    def __init__(self, id=None, name=None, is_active=None, username=None,
+                 password=None, proto_version=None, sig_method=None,
+                 max_nonce_log=None, cluster=None):
+        self.id = id
+        self.name = name
+        self.is_active = is_active
+        self.username = username
+        self.password = password
+        self.proto_version = proto_version
+        self.sig_method = sig_method
+        self.max_nonce_log = max_nonce_log
+        self.cluster = cluster
+
     def to_json(self):
         return to_json(self)
 
@@ -250,37 +278,65 @@ class HTTPSOAP(Base):
     __table_args__ = (UniqueConstraint('name', 'connection', 'transport', 'cluster_id'),
                       UniqueConstraint('url_path', 'connection', 'soap_action', 'cluster_id'), {})
 
-    id = Column(Integer,  Sequence('http_soap_seq'), primary_key=True)
+    id = Column(Integer, Sequence('http_soap_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
     is_internal = Column(Boolean(), nullable=False)
 
     connection = Column(Enum('channel', 'outgoing', name='http_soap_connection'), nullable=False)
     transport = Column(Enum('plain_http', 'soap', name='http_soap_transport'), nullable=False)
-    
+
     host = Column(String(200), nullable=True)
     url_path = Column(String(200), nullable=False)
     method = Column(String(200), nullable=True)
 
     soap_action = Column(String(200), nullable=False)
     soap_version = Column(String(20), nullable=True)
-    
+
     data_format = Column(String(20), nullable=True)
+
+    # New in 1.2
+    ping_method = Column(String(60), nullable=True)
+
+    # New in 1.2
+    pool_size = Column(Integer, nullable=True)
+
+    # New in 1.2
+    merge_url_params_req = Column(Boolean, nullable=True, default=True)
+
+    # New in 1.2
+    url_params_pri = Column(String(200), nullable=True, default='path-over-qs')
+
+    # New in 1.2
+    params_pri = Column(String(200), nullable=True, default='channel-params-over-msg')
     
+    # New in 1.2
+    audit_enabled = Column(Boolean, nullable=False, default=False)
+    
+    # New in 1.2
+    audit_back_log = Column(Integer, nullable=False, default=MISC.DEFAULT_AUDIT_BACK_LOG)
+    
+    # New in 1.2
+    audit_max_payload = Column(Integer, nullable=False, default=MISC.DEFAULT_AUDIT_MAX_PAYLOAD)
+    
+    # New in 1.2
+    audit_repl_patt_type = Column(String(), nullable=False, default=MSG_PATTERN_TYPE.ELEM_PATH.id)
+
     service_id = Column(Integer, ForeignKey('service.id', ondelete='CASCADE'), nullable=True)
     service = relationship('Service', backref=backref('http_soap', order_by=name, cascade='all, delete, delete-orphan'))
-    
+
     cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
     cluster = relationship(Cluster, backref=backref('http_soap_list', order_by=name, cascade='all, delete, delete-orphan'))
-    
+
     security_id = Column(Integer, ForeignKey('sec_base.id', ondelete='CASCADE'), nullable=True)
     security = relationship(SecurityBase, backref=backref('http_soap_list', order_by=name, cascade='all, delete, delete-orphan'))
-    
-    def __init__(self, id=None, name=None, is_active=None, is_internal=None, 
-                 connection=None, transport=None, host=None, url_path=None, method=None, 
-                 soap_action=None, soap_version=None, data_format=None, service_id=None, service=None,
-                 security=None, cluster_id=None, cluster=None, service_name=None,
-                 security_id=None, security_name=None):
+
+    def __init__(self, id=None, name=None, is_active=None, is_internal=None,
+                 connection=None, transport=None, host=None, url_path=None, method=None,
+                 soap_action=None, soap_version=None, data_format=None, ping_method=None,
+                 pool_size=None, merge_url_params_req=None, url_params_pri=None,
+                 params_pri=None, service_id=None, service=None, security=None, cluster_id=None,
+                 cluster=None, service_name=None, security_id=None, security_name=None):
         self.id = id
         self.name = name
         self.is_active = is_active
@@ -293,6 +349,11 @@ class HTTPSOAP(Base):
         self.soap_action = soap_action
         self.soap_version = soap_version
         self.data_format = data_format
+        self.ping_method = ping_method
+        self.pool_size = pool_size
+        self.merge_url_params_req = merge_url_params_req
+        self.url_params_pri = url_params_pri
+        self.params_pri = params_pri
         self.service_id = service_id
         self.service = service
         self.security = security
@@ -310,7 +371,7 @@ class SQLConnectionPool(Base):
     __tablename__ = 'sql_pool'
     __table_args__ = (UniqueConstraint('cluster_id', 'name'), {})
 
-    id = Column(Integer,  Sequence('sql_pool_id_seq'), primary_key=True)
+    id = Column(Integer, Sequence('sql_pool_id_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
     username = Column(String(200), nullable=False)
@@ -324,11 +385,11 @@ class SQLConnectionPool(Base):
 
     cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
     cluster = relationship(Cluster, backref=backref('sql_pools', order_by=name, cascade='all, delete, delete-orphan'))
-    
+
     engine_text = None # For auto-completion, not used by the DB
 
-    def __init__(self, id=None, name=None, is_active=None, db_name=None, 
-                 username=None, engine=None, extra=None, host=None, port=None, 
+    def __init__(self, id=None, name=None, is_active=None, db_name=None,
+                 username=None, engine=None, extra=None, host=None, port=None,
                  pool_size=None, cluster=None):
         self.id = id
         self.name = name
@@ -350,20 +411,20 @@ class Service(Base):
     __tablename__ = 'service'
     __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
 
-    id = Column(Integer,  Sequence('service_id_seq'), primary_key=True)
-    name = Column(String(2000), nullable=False)
+    id = Column(Integer, Sequence('service_id_seq'), primary_key=True)
+    name = Column(String(300), nullable=False)
     is_active = Column(Boolean(), nullable=False)
     impl_name = Column(String(2000), nullable=False)
     is_internal = Column(Boolean(), nullable=False)
     wsdl = Column(LargeBinary(5000000), nullable=True)
     wsdl_name = Column(String(200), nullable=True)
-    
-    slow_threshold = Column(Integer,  nullable=False, default=99999)
-    
+
+    slow_threshold = Column(Integer, nullable=False, default=99999)
+
     cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
     cluster = relationship(Cluster, backref=backref('services', order_by=name, cascade='all, delete, delete-orphan'))
 
-    def __init__(self, id=None, name=None, is_active=None, impl_name=None, 
+    def __init__(self, id=None, name=None, is_active=None, impl_name=None,
                  is_internal=None, cluster=None, wsdl=None, wsdl_name=None):
         self.id = id
         self.name = name
@@ -391,14 +452,14 @@ class Service(Base):
         self.sample_req_resp_freq = None # Not used by the database
         self.sample_req_html = None # Not used by the database
         self.sample_resp_html = None # Not used by the database
-        
+
         self.usage = None # Not used by the database
         self.time_last = None # Not used by the database
-        
+
         self.time_min_all_time = None # Not used by the database
         self.time_max_all_time = None # Not used by the database
         self.time_mean_all_time = None # Not used by the database
-        
+
         self.time_usage_1h = None # Not used by the database
         self.time_min_1h = None # Not used by the database
         self.time_max_1h = None # Not used by the database
@@ -444,7 +505,7 @@ class Job(Base):
     __tablename__ = 'job'
     __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
 
-    id = Column(Integer,  Sequence('job_id_seq'), primary_key=True)
+    id = Column(Integer, Sequence('job_id_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
     job_type = Column(Enum(SCHEDULER_JOB_TYPE.ONE_TIME, SCHEDULER_JOB_TYPE.INTERVAL_BASED,
@@ -484,7 +545,7 @@ class IntervalBasedJob(Base):
     __tablename__ = 'job_interval_based'
     __table_args__ = (UniqueConstraint('job_id'), {})
 
-    id = Column(Integer,  Sequence('job_intrvl_seq'), primary_key=True)
+    id = Column(Integer, Sequence('job_intrvl_seq'), primary_key=True)
     job_id = Column(Integer, nullable=False)
 
     weeks = Column(Integer, nullable=True)
@@ -515,7 +576,7 @@ class CronStyleJob(Base):
     __tablename__ = 'job_cron_style'
     __table_args__ = (UniqueConstraint('job_id'), {})
 
-    id = Column(Integer,  Sequence('job_cron_seq'), primary_key=True)
+    id = Column(Integer, Sequence('job_cron_seq'), primary_key=True)
     cron_definition = Column(String(4000), nullable=False)
 
     job_id = Column(Integer, ForeignKey('job.id', ondelete='CASCADE'), nullable=False)
@@ -534,10 +595,10 @@ class ConnDefAMQP(Base):
     __tablename__ = 'conn_def_amqp'
     __table_args__ = (UniqueConstraint('name', 'cluster_id', 'def_type'), {})
 
-    id = Column(Integer,  Sequence('conn_def_amqp_seq'), primary_key=True)
+    id = Column(Integer, Sequence('conn_def_amqp_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     # TODO is_active = Column(Boolean(), nullable=False)
-    
+
     def_type = Column(String(10), nullable=False)
     host = Column(String(200), nullable=False)
     port = Column(Integer(), nullable=False)
@@ -551,7 +612,7 @@ class ConnDefAMQP(Base):
     cluster = relationship(Cluster, backref=backref('amqp_conn_defs', order_by=name, cascade='all, delete, delete-orphan'))
 
     def __init__(self, id=None, name=None, def_type=None, host=None, port=None,
-                 vhost=None,  username=None,  password=None, frame_max=None,
+                 vhost=None, username=None, password=None, frame_max=None,
                  heartbeat=None, cluster_id=None):
         self.id = id
         self.name = name
@@ -571,7 +632,7 @@ class ConnDefWMQ(Base):
     __tablename__ = 'conn_def_wmq'
     __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
 
-    id = Column(Integer,  Sequence('conn_def_wmq_seq'), primary_key=True)
+    id = Column(Integer, Sequence('conn_def_wmq_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     # TODO is_active = Column(Boolean(), nullable=False)
 
@@ -594,7 +655,7 @@ class ConnDefWMQ(Base):
 
     def __init__(self, id=None, name=None, host=None, port=None,
                  queue_manager=None, channel=None, cache_open_send_queues=None,
-                 cache_open_receive_queues=None,  use_shared_connections=None, ssl=None,
+                 cache_open_receive_queues=None, use_shared_connections=None, ssl=None,
                  ssl_cipher_spec=None, ssl_key_repository=None, needs_mcd=None,
                  max_chars_printed=None, cluster_id=None):
         self.id = id
@@ -621,7 +682,7 @@ class OutgoingAMQP(Base):
     __tablename__ = 'out_amqp'
     __table_args__ = (UniqueConstraint('name', 'def_id'), {})
 
-    id = Column(Integer,  Sequence('out_amqp_seq'), primary_key=True)
+    id = Column(Integer, Sequence('out_amqp_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
 
@@ -654,14 +715,14 @@ class OutgoingAMQP(Base):
         self.def_id = def_id
         self.delivery_mode_text = delivery_mode_text # Not used by the DB
         self.def_name = def_name # Not used by the DB
-        
+
 class OutgoingFTP(Base):
     """ An outgoing FTP connection.
     """
     __tablename__ = 'out_ftp'
     __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
 
-    id = Column(Integer,  Sequence('out_ftp_seq'), primary_key=True)
+    id = Column(Integer, Sequence('out_ftp_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
 
@@ -676,7 +737,7 @@ class OutgoingFTP(Base):
     cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
     cluster = relationship(Cluster, backref=backref('out_conns_ftp', order_by=name, cascade='all, delete, delete-orphan'))
 
-    def __init__(self, id=None, name=None, is_active=None, host=None, user=None, 
+    def __init__(self, id=None, name=None, is_active=None, host=None, user=None,
                  password=None, acct=None, timeout=None, port=None, dircache=None,
                  cluster_id=None):
         self.id = id
@@ -697,7 +758,7 @@ class OutgoingWMQ(Base):
     __tablename__ = 'out_wmq'
     __table_args__ = (UniqueConstraint('name', 'def_id'), {})
 
-    id = Column(Integer,  Sequence('out_wmq_seq'), primary_key=True)
+    id = Column(Integer, Sequence('out_wmq_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
 
@@ -727,7 +788,7 @@ class OutgoingZMQ(Base):
     __tablename__ = 'out_zmq'
     __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
 
-    id = Column(Integer,  Sequence('out_zmq_seq'), primary_key=True)
+    id = Column(Integer, Sequence('out_zmq_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
 
@@ -754,7 +815,7 @@ class ChannelAMQP(Base):
     __tablename__ = 'channel_amqp'
     __table_args__ = (UniqueConstraint('name', 'def_id'), {})
 
-    id = Column(Integer,  Sequence('channel_amqp_seq'), primary_key=True)
+    id = Column(Integer, Sequence('channel_amqp_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
     queue = Column(String(200), nullable=False)
@@ -786,7 +847,7 @@ class ChannelWMQ(Base):
     __tablename__ = 'channel_wmq'
     __table_args__ = (UniqueConstraint('name', 'def_id'), {})
 
-    id = Column(Integer,  Sequence('channel_wmq_seq'), primary_key=True)
+    id = Column(Integer, Sequence('channel_wmq_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
     queue = Column(String(200), nullable=False)
@@ -815,7 +876,7 @@ class ChannelZMQ(Base):
     __tablename__ = 'channel_zmq'
     __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
 
-    id = Column(Integer,  Sequence('channel_zmq_seq'), primary_key=True)
+    id = Column(Integer, Sequence('channel_zmq_seq'), primary_key=True)
     name = Column(String(200), nullable=False)
     is_active = Column(Boolean(), nullable=False)
 
@@ -823,7 +884,7 @@ class ChannelZMQ(Base):
     socket_type = Column(String(20), nullable=False)
     sub_key = Column(String(200), nullable=True)
     data_format = Column(String(20), nullable=True)
-    
+
     service_id = Column(Integer, ForeignKey('service.id', ondelete='CASCADE'), nullable=False)
     service = relationship(Service, backref=backref('channels_zmq', order_by=name, cascade='all, delete, delete-orphan'))
 
@@ -846,44 +907,287 @@ class DeploymentPackage(Base):
     a Distutils2 archive.
     """
     __tablename__ = 'deployment_package'
-    
-    id = Column(Integer,  Sequence('depl_package_seq'), primary_key=True)
+
+    id = Column(Integer, Sequence('depl_package_seq'), primary_key=True)
     deployment_time = Column(DateTime(), nullable=False)
     details = Column(String(2000), nullable=False)
-    
+
     payload_name = Column(String(200), nullable=False)
     payload = Column(LargeBinary(5000000), nullable=False)
-    
+
     server_id = Column(Integer, ForeignKey('server.id', ondelete='CASCADE'), nullable=False, primary_key=False)
     server = relationship(Server, backref=backref('originating_deployment_packages', order_by=deployment_time, cascade='all, delete, delete-orphan'))
-    
+
     def __init__(self, id=None, deployment_time=None, details=None, payload_name=None, payload=None):
         self.id = id
         self.deployment_time = deployment_time
         self.details = details
         self.payload_name = payload_name
         self.payload = payload
-    
+
 class DeploymentStatus(Base):
     """ Whether a server has already deployed a given package.
     """
     __tablename__ = 'deployment_status'
     __table_args__ = (UniqueConstraint('package_id', 'server_id'), {})
 
-    id = Column(Integer,  Sequence('depl_status_seq'), primary_key=True)    
-    
+    id = Column(Integer, Sequence('depl_status_seq'), primary_key=True)
+
     package_id = Column(Integer, ForeignKey('deployment_package.id', ondelete='CASCADE'), nullable=False, primary_key=False)
     package = relationship(DeploymentPackage, backref=backref('deployment_status_list', order_by=package_id, cascade='all, delete, delete-orphan'))
-    
+
     server_id = Column(Integer, ForeignKey('server.id', ondelete='CASCADE'), nullable=False, primary_key=False)
     server = relationship(Server, backref=backref('deployment_status_list', order_by=server_id, cascade='all, delete, delete-orphan'))
 
     # See zato.common.DEPLOYMENT_STATUS
     status = Column(String(20), nullable=False)
     status_change_time = Column(DateTime(), nullable=False)
-    
+
     def __init__(self, package_id=None, server_id=None, status=None, status_change_time=None):
         self.package_id = package_id
         self.server_id = server_id
         self.status = status
         self.status_change_time = status_change_time
+
+################################################################################
+
+class DeliveryDefinitionBase(Base):
+    """ A guaranteed delivery's definition (base class).
+    """
+    __tablename__ = 'delivery_def_base'
+    __mapper_args__ = {'polymorphic_on': 'target_type'}
+
+    id = Column(Integer, Sequence('deliv_def_seq'), primary_key=True)
+    name = Column(String(200), nullable=False, index=True)
+    short_def = Column(String(200), nullable=False)
+    last_used = Column(DateTime(), nullable=True)
+
+    target_type = Column(String(200), nullable=False)
+    callback_list = Column(LargeBinary(10000), nullable=True)
+
+    expire_after = Column(Integer, nullable=False)
+    expire_arch_succ_after = Column(Integer, nullable=False)
+    expire_arch_fail_after = Column(Integer, nullable=False)
+    check_after = Column(Integer, nullable=False)
+    retry_repeats = Column(Integer, nullable=False)
+    retry_seconds = Column(Integer, nullable=False)
+
+    cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
+    cluster = relationship(Cluster, backref=backref('delivery_list', order_by=name, cascade='all, delete, delete-orphan'))
+
+class DeliveryDefinitionOutconnWMQ(DeliveryDefinitionBase):
+    """ A guaranteed delivery's definition (outgoing WebSphere MQ connections).
+    """
+    __tablename__ = 'delivery_def_out_wmq'
+    __mapper_args__ = {'polymorphic_identity': INVOCATION_TARGET.OUTCONN_WMQ}
+
+    id = Column(Integer, ForeignKey('delivery_def_base.id'), primary_key=True)
+    target_id = Column(Integer, ForeignKey('out_wmq.id', ondelete='CASCADE'), nullable=False, primary_key=False)
+    target = relationship(OutgoingWMQ, backref=backref('delivery_def_list', order_by=target_id, cascade='all, delete, delete-orphan'))
+
+    def __init__(self, id=None, target_id=None):
+        self.id = id
+        self.target_id = target_id
+
+class Delivery(Base):
+    """ A guaranteed delivery.
+    """
+    __tablename__ = 'delivery'
+
+    id = Column(Integer, Sequence('deliv_seq'), primary_key=True)
+    task_id = Column(String(64), unique=True, nullable=False, index=True)
+
+    name = Column(String(200), nullable=False)
+    creation_time = Column(DateTime(), nullable=False)
+
+    args = Column(LargeBinary(1000000), nullable=True)
+    kwargs = Column(LargeBinary(1000000), nullable=True)
+
+    last_used = Column(DateTime(), nullable=True)
+    resubmit_count = Column(Integer, nullable=False, default=0)
+
+    state = Column(String(200), nullable=False, index=True)
+
+    source_count = Column(Integer, nullable=False, default=1)
+    target_count = Column(Integer, nullable=False, default=0)
+
+    definition_id = Column(Integer, ForeignKey('delivery_def_base.id', ondelete='CASCADE'), nullable=False, primary_key=False)
+    definition = relationship(DeliveryDefinitionBase, backref=backref('delivery_list', order_by=creation_time, cascade='all, delete, delete-orphan'))
+
+class DeliveryPayload(Base):
+    """ A guaranteed delivery's payload.
+    """
+    __tablename__ = 'delivery_payload'
+
+    id = Column(Integer, Sequence('deliv_payl_seq'), primary_key=True)
+    task_id = Column(String(64), unique=True, nullable=False, index=True)
+
+    creation_time = Column(DateTime(), nullable=False)
+    payload = Column(LargeBinary(5000000), nullable=False)
+
+    delivery_id = Column(Integer, ForeignKey('delivery.id', ondelete='CASCADE'), nullable=False, primary_key=False)
+    delivery = relationship(Delivery, backref=backref('payload', uselist=False, cascade='all, delete, delete-orphan', single_parent=True))
+
+class DeliveryHistory(Base):
+    """ A guaranteed delivery's history.
+    """
+    __tablename__ = 'delivery_history'
+
+    id = Column(Integer, Sequence('deliv_payl_seq'), primary_key=True)
+    task_id = Column(String(64), nullable=False, index=True)
+
+    entry_type = Column(String(64), nullable=False)
+    entry_time = Column(DateTime(), nullable=False, index=True)
+    entry_ctx = Column(LargeBinary(6000000), nullable=False)
+    resubmit_count = Column(Integer, nullable=False, default=0) # Copy of delivery.resubmit_count so it's known for each history entry
+
+    delivery_id = Column(Integer, ForeignKey('delivery.id', ondelete='CASCADE'), nullable=False, primary_key=False)
+    delivery = relationship(Delivery, backref=backref('history_list', order_by=entry_time, cascade='all, delete, delete-orphan'))
+
+# ##############################################################################
+
+class MsgNamespace(Base):
+    """ A message namespace, used in XPath, for instance.
+    """
+    __tablename__ = 'msg_ns'
+    __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
+
+    id = Column(Integer, Sequence('msg_ns_seq'), primary_key=True)
+    name = Column(String(200), nullable=False)
+    value = Column(String(500), nullable=False)
+
+    cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
+    cluster = relationship(Cluster, backref=backref('namespaces', order_by=name, cascade='all, delete, delete-orphan'))
+
+    def __init__(self, id=None, name=None, value=None, cluster_id=None):
+        self.id = id
+        self.name = name
+        self.value = value
+        self.cluster_id = cluster_id
+
+class XPath(Base):
+    """ An XPath expression to run against XML messages.
+    """
+    __tablename__ = 'msg_xpath'
+    __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
+
+    id = Column(Integer, Sequence('msg_xpath_seq'), primary_key=True)
+    name = Column(String(200), nullable=False)
+    value = Column(String(1500), nullable=False)
+
+    cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
+    cluster = relationship(Cluster, backref=backref('xpaths', order_by=name, cascade='all, delete, delete-orphan'))
+
+    def __init__(self, id=None, name=None, value=None, cluster_id=None):
+        self.id = id
+        self.name = name
+        self.value = value
+        self.cluster_id = cluster_id
+
+class ElemPath(Base):
+    """ An XPath-list expression to run against JSON messages.
+    """
+    __tablename__ = 'msg_elem_path'
+    __table_args__ = (UniqueConstraint('name', 'cluster_id'), {})
+
+    id = Column(Integer, Sequence('msg_elem_path_seq'), primary_key=True)
+    name = Column(String(200), nullable=False)
+    value = Column(String(1500), nullable=False)
+
+    cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
+    cluster = relationship(Cluster, backref=backref('elem_paths', order_by=name, cascade='all, delete, delete-orphan'))
+
+    def __init__(self, id=None, name=None, value=None, cluster_id=None):
+        self.id = id
+        self.name = name
+        self.value = value
+        self.cluster_id = cluster_id
+
+# ##############################################################################
+
+class HTTSOAPAudit(Base):
+    """ An audit log for HTTP/SOAP channels and outgoing connections.
+    """
+    __tablename__ = 'http_soap_audit'
+
+    id = Column(Integer, Sequence('http_soap_audit_seq'), primary_key=True)
+    name = Column(String(), nullable=False, index=True)
+    cid = Column(String(), nullable=False, index=True)
+    
+    transport = Column(String(), nullable=False, index=True)
+    connection = Column(String(), nullable=False, index=True)
+    
+    req_time = Column(DateTime(), nullable=False)
+    resp_time = Column(DateTime(), nullable=True)
+    
+    user_token = Column(String(), nullable=True, index=True)
+    invoke_ok = Column(Boolean(), nullable=True)
+    auth_ok = Column(Boolean(), nullable=True)
+    remote_addr = Column(String(), nullable=False, index=True)
+    
+    req_headers = Column(String(), nullable=True)
+    req_payload = Column(String(), nullable=True)
+    resp_headers = Column(String(), nullable=True)
+    resp_payload = Column(String(), nullable=True)
+
+    cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
+    conn_id = Column(Integer, ForeignKey('http_soap.id', ondelete='CASCADE'), nullable=False)
+
+    def __init__(self, id=None, name=None, cid=None, transport=None, 
+            connection=None, req_time=None, resp_time=None, user_token=None, 
+            invoke_ok=None, auth_ok=None, remote_addr=None, req_headers=None,
+            req_payload=None, resp_headers=None, resp_payload=None):
+        
+        self.id = id
+        self.name = name
+        self.cid = cid
+        
+        self.transport = transport
+        self.connection = connection
+        
+        self.req_time = req_time
+        self.resp_time = resp_time
+        
+        self.user_token = user_token
+        self.invoke_ok = invoke_ok
+        self.auth_ok = auth_ok
+        self.remote_addr = remote_addr
+        
+        self.req_headers = req_headers
+        self.req_payload = req_payload
+        self.resp_headers = resp_headers
+        self.resp_payload = resp_payload
+
+class HTTSOAPAuditReplacePatternsElemPath(Base):
+    """ ElemPath replace patterns for HTTP/SOAP connections.
+    """
+    __tablename__ = 'http_soap_au_rpl_p_ep'
+    __table_args__ = (UniqueConstraint('conn_id', 'pattern_id'), {})
+    
+    id = Column(Integer, Sequence('htp_sp_ad_rpl_p_ep_seq'), primary_key=True)
+    conn_id = Column(Integer, ForeignKey('http_soap.id', ondelete='CASCADE'), nullable=False)
+    pattern_id = Column(Integer, ForeignKey('msg_elem_path.id', ondelete='CASCADE'), nullable=False)
+    cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
+    
+    replace_patterns_elem_path = relationship(HTTPSOAP, 
+        backref=backref('replace_patterns_elem_path', order_by=id, cascade='all, delete, delete-orphan'))
+
+    pattern = relationship(ElemPath)
+    
+class HTTSOAPAuditReplacePatternsXPath(Base):
+    """ XPath replace patterns for HTTP/SOAP connections.
+    """
+    __tablename__ = 'http_soap_au_rpl_p_xp'
+    __table_args__ = (UniqueConstraint('conn_id', 'pattern_id'), {})
+    
+    id = Column(Integer, Sequence('htp_sp_ad_rpl_p_xp_seq'), primary_key=True)
+    conn_id = Column(Integer, ForeignKey('http_soap.id', ondelete='CASCADE'), nullable=False)
+    pattern_id = Column(Integer, ForeignKey('msg_xpath.id', ondelete='CASCADE'), nullable=False)
+    cluster_id = Column(Integer, ForeignKey('cluster.id', ondelete='CASCADE'), nullable=False)
+    
+    replace_patterns_xpath = relationship(HTTPSOAP, 
+        backref=backref('replace_patterns_xpath', order_by=id, cascade='all, delete, delete-orphan'))
+
+    pattern = relationship(XPath)
+
+# ##############################################################################
